import 'dart:mirrors';

import 'package:xdatabase/src/table_meta.dart';
import 'package:xdatabase/src/xdatabase_base.dart';

import 'connection_pool.dart';

class NotSupportedException extends Error {}

typedef HandleObjectMap = Map<String, dynamic> Function(dynamic obj);
typedef HandleMapObject = dynamic Function(Map<String, dynamic> row);

class Query<T> {
  QueryAble ctx;
  String statement;
  HandleObjectMap objectMap;
  HandleMapObject mapObject;
  late TableDeclaration tableInfo;
  Query(this.statement, this.ctx, this.objectMap, this.mapObject) {
    tableInfo = DeclarationHolder().ofTable<T>()!;
  }

  String trimSql() {
    final idx = statement.toLowerCase().indexOf('where');
    return statement.substring(idx);
  }

  String _deleteSql() {
    return 'DELETE FROM "${tableInfo.tableName}" $statement';
  }

  String _updateSql() {
    final lines = <String>[
      'UPDATE "${tableInfo.tableName}" SET ',
      tableInfo.fields
          .map((FieldDeclaration fd) => '"${fd.fieldName}"=@${fd.fieldName}')
          .toList()
          .join(', '),
      ' ${trimSql()};'
    ];
    return lines.join('');
  }

  Future<int> update(T obj) async {
    final sql = _updateSql();
    final params = objectMap(obj);
    return await ctx.execute(sql, substitutionValues: params);
  }

  Future<int> delete(T obj) async {
    final sql = _deleteSql();
    final params = objectMap(obj);
    return await ctx.execute(sql, substitutionValues: params);
  }

  Future<List<T>> fetch(T? obj, {int limit = 0, int offset = 0}) async {
    List<Map<String, Map<String, dynamic>>> resultSet;
    if (obj != null) {
      final params = objectMap(obj);
      resultSet =
          await ctx.mappedResultsQuery(statement, substitutionValues: params);
    } else {
      resultSet = await ctx.mappedResultsQuery(statement);
    }
    if (resultSet.isNotEmpty) {
      return resultSet
          .map((row) => mapObject(row[tableInfo.tableName]!) as T)
          .toList();
    } else {
      return <T>[];
    }
  }
}

class Orm<T> {
  late TableDeclaration tableInfo;
  QueryAble ctx;
  Orm(this.ctx) {
    tableInfo = DeclarationHolder().ofTable<T>()!;
  }

  String _insertSQL() {
    var lines = [
      'INSERT INTO ${tableInfo.tableName}("',
      tableInfo.fields.map((fd) => fd.fieldName).toList().join('","'),
      '") VALUES (',
      tableInfo.fields.map((fd) => '@${fd.fieldName}').toList().join(','),
      ');'
    ];
    return lines.join('');
  }

  Map<String, dynamic> objectMap(dynamic obj) {
    final data = <String, dynamic>{};
    final om = reflect(obj);
    tableInfo.fields.forEach((fd) {
      final value = om.getField(fd.symbol);
      data.addEntries([MapEntry(fd.fieldName, value.reflectee)]);
    });
    return data;
  }

  dynamic mapObject(Map<String, dynamic> row) {
    final cm = reflectClass(T);
    final ins = cm.newInstance(const Symbol(''), [], {});
    tableInfo.fields.forEach((fd) {
      //params.addEntries([MapEntry(fd.symbol, row[fd.fieldName])]);
      ins.setField(fd.symbol, row[fd.fieldName]);
    });
    return ins.reflectee;
  }

  Future<int> insert(T obj) async {
    final sql = _insertSQL();
    final param = objectMap(obj);
    print(param);
    return await ctx.execute(sql, substitutionValues: param);
  }

  Query<T> query(String statement) {
    return Query(statement.replaceAll('[table]', tableInfo.tableName), ctx,
        objectMap, mapObject);
  }

  Future<T?> byPrimary(T obj) async {
    final pks = tableInfo.fields.where((element) => element.primary).toList();
    if (pks.isEmpty) {
      throw NotSupportedException();
    }
    final sql =
        'SELECT * FROM "${tableInfo.tableName}" WHERE "${pks.first.fieldName}"=@${pks.first.fieldName};';
    final query = Query<T>(sql, ctx, objectMap, objectMap);
    final results = await query.fetch(obj);
    if (results.isNotEmpty) {
      return results.first;
    }
  }
}
